# Copyright 2024 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Auxiliary functions for knots calculations."""

import bisect
from collections.abc import Collection, Sequence
import dataclasses
import numpy as np


__all__ = [
    'KnotInfo',
    'get_knot_info',
    'l1_distance_weights',
]


# TODO: Reimplement with a more readable method.
def _find_neighboring_knots_indices(
    times: np.ndarray,
    knot_locations: np.ndarray,
) -> Sequence[Sequence[int] | None]:
  """Return indices of neighboring knot locations.

  Returns indices in `knot_locations` that correspond to the neighboring knot
  locations for each time period. If a time point is at or before the first
  knot, the first knot is the only neighboring knot. If a time point is after
  the last knot, the last knot is the only neighboring knot.

  Args:
    times: Times `0, 1, 2,..., (n_times-1)`.
    knot_locations: The location of knots within `0, 1, 2,..., (n_times-1)`.

  Returns:
    List of length `n_times`. Each element is the indices of the neighboring
    knot locations for the respective time period. If a time point is at or
    before the first knot, the first knot is the only neighboring knot. If a
    time point is after the last knot, the last knot is the only neighboring
    knot.
  """
  neighboring_knots_indices = [None] * len(times)
  for t in times:
    # knot_locations assumed to be sorted.
    if t <= knot_locations[0]:
      neighboring_knots_indices[t] = [0]
    elif t >= knot_locations[-1]:
      neighboring_knots_indices[t] = [len(knot_locations) - 1]
    else:
      bisect_index = bisect.bisect_left(knot_locations, t)
      neighboring_knots_indices[t] = [bisect_index - 1, bisect_index]
  return neighboring_knots_indices


def l1_distance_weights(
    n_times: int, knot_locations: np.ndarray[int, np.dtype[int]]
) -> np.ndarray:
  """Computes weights at knots for every time period.

  The two neighboring knots inform the weight estimate of a particular time
  period. The amount each of the two neighboring knots inform the weight at a
  time period depends on how close (L1 distance) they are to the time period. If
  a time point coincides with a knot location, then 100% weight is given to that
  knot. If a time point lies outside the range of knots, then 100% weight is
  given to the nearest endpoint knot.

  This function computes an `(n_knots, n_times)` array of weights that are used
  to model trend and seasonality. For a given time, the array contains two
  non-zero weights. The weights are inversely proportional to the L1 distance
  from the given time to the neighboring knots. The two weights are normalized
  such that they sum to 1.

  Args:
    n_times: The number of time points.
    knot_locations: The location of knots within `0, 1, 2,..., (n_times-1)`.

  Returns:
    A weight array with dimensions `(n_knots, n_times)` with values summing up
    to 1 for each time period when summing over knots.
  """
  if knot_locations.ndim != 1:
    raise ValueError('`knot_locations` must be one-dimensional.')
  if not np.all(knot_locations == np.sort(knot_locations)):
    raise ValueError('`knot_locations` must be sorted.')
  if len(knot_locations) <= 1:
    raise ValueError('Number of knots must be greater than 1.')
  if len(knot_locations) != len(np.unique(knot_locations)):
    raise ValueError('`knot_locations` must be unique.')
  if np.any(knot_locations < 0):
    raise ValueError('knot_locations must be positive.')
  if np.any(knot_locations >= n_times):
    raise ValueError('knot_locations must be less than `n_times`.')

  times = np.arange(n_times)
  time_minus_knot = abs(knot_locations[:, np.newaxis] - times[np.newaxis, :])

  w = np.zeros(time_minus_knot.shape, dtype=np.float32)
  neighboring_knots_indices = _find_neighboring_knots_indices(
      times, knot_locations
  )
  for t in times:
    idx = neighboring_knots_indices[t]
    if len(idx) == 1:
      w[idx, t] = 1
    else:
      # Weight is in proportion to how close the two neighboring knots are.
      w[idx, t] = 1 - (time_minus_knot[idx, t] / time_minus_knot[idx, t].sum())

  return w


def _get_equally_spaced_knot_locations(n_times, n_knots):
  """Equally spaced knot locations starting at the endpoints."""
  return np.linspace(0, n_times - 1, n_knots, dtype=int)


@dataclasses.dataclass(frozen=True)
class KnotInfo:
  """Contains the number of knots, knot locations, and weights.

  Attributes:
    n_knots: The number of knots
    knot_locations: The location of knots
    weights: The weights used to multiply with the knot values to get time-
      varying coefficients.
  """

  n_knots: int
  knot_locations: np.ndarray[int, np.dtype[int]]
  weights: np.ndarray[int, np.dtype[float]]


def get_knot_info(
    n_times: int,
    knots: int | Collection[int] | None,
    is_national: bool = False,
) -> KnotInfo:
  """Returns the number of knots, knot locations, and weights.

  Args:
    n_times: The number of time periods in the data.
    knots: An optional integer or a collection of integers indicating the knots
      used to estimate time effects. When `knots` is a collection of integers,
      the knot locations are provided by that collection. Zero corresponds to a
      knot at the first time period, one corresponds to a knot at the second
      time, ..., and `(n_times - 1)` corresponds to a knot at the last time
      period. When `knots` is an integer, then there are knots with locations
      equally spaced across the time periods (including knots at zero and
      `(n_times - 1)`. When `knots` is `1`, there is a single common regression
      coefficient used for all time periods. If `knots` is `None`, then the
      numbers of knots used is equal to the number of time periods. This is
      equivalent to each time period having its own regression coefficient.
    is_national: A boolean indicator whether to adapt the knot information for a
      national model.

  Returns:
    A KnotInfo that contains the number of knots, the location of knots, and the
    weights used to multiply with the knot values to get time-varying
    coefficients.
  """

  if isinstance(knots, int):
    if knots < 1:
      raise ValueError('If knots is an integer, it must be at least 1.')
    elif knots > n_times:
      raise ValueError(
          f'The number of knots ({knots}) cannot be greater than the number of'
          f' time periods in the kpi ({n_times}).'
      )
    elif is_national and knots == n_times:
      raise ValueError(
          f'Number of knots ({knots}) must be less than number of time periods'
          f' ({n_times}) in a nationally aggregated model.'
      )
    n_knots = knots
    knot_locations = _get_equally_spaced_knot_locations(n_times, n_knots)
  elif isinstance(knots, Collection) and knots:
    if any(k < 0 for k in knots):
      raise ValueError('Knots must be all non-negative.')
    if any(k >= n_times for k in knots):
      raise ValueError(
          'Knots must all be less than the number of time periods.'
      )
    n_knots = len(knots)
    # np.unique also sorts
    knot_locations = np.unique(knots)
  elif isinstance(knots, Collection):
    raise ValueError('Knots cannot be empty.')
  else:
    # knots is None
    n_knots = 1 if is_national else n_times
    knot_locations = _get_equally_spaced_knot_locations(n_times, n_knots)

  if n_knots == 1:
    weights = np.ones((1, n_times), dtype=np.float32)
  else:
    weights = l1_distance_weights(n_times, knot_locations)

  return KnotInfo(n_knots, knot_locations, weights)



# Copyright 2024 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The setup.py file for Meridian."""

from pathlib import Path
import sass
from setuptools import Command, setup
from setuptools.command.build import build


class ScssCompileCommand(Command):
  _templates_path = 'meridian/analysis/templates'
  _scss_output_mapping = {
      'style.scss': 'style.css',
  }

  def initialize_options(self):
    self.build_lib = None

  def finalize_options(self):
    self.set_undefined_options('build_py', ('build_lib', 'build_lib'))

  def run(self):
    if not self.build_lib:
      return

    for scss, css in self._scss_output_mapping.items():
      self._compile_scss(scss, css)

  def _compile_scss(self, source_filename, dest_filename):
    dirpath = Path(self.build_lib).joinpath(self._templates_path)
    dirpath.mkdir(parents=True, exist_ok=True)
    dest_filepath = dirpath.joinpath(dest_filename)

    source_relpath = self._templates_path + '/' + source_filename

    print(f'compiling {source_relpath} -> {dest_filepath}')

    with open(Path.cwd().joinpath(source_relpath), 'r') as f:
      css = sass.compile(string=f.read())

    dest_filepath.write_text(css, encoding='utf-8')

  def get_outputs(self):
    dirpath = Path(self.build_lib).joinpath(self._templates_path)
    return [
        dirpath.joinpath(css) for _, css in self._scss_output_mapping.items()
    ]


class CustomBuild(build):
  sub_commands = [('compile_scss', None)] + build.sub_commands


if __name__ == '__main__':
  setup(cmdclass={'build': CustomBuild, 'compile_scss': ScssCompileCommand})



# Copyright 2024 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Structures and functions for manipulating media value data and tensors."""

import dataclasses
from meridian import constants
from meridian.data import input_data as data
from meridian.model import spec
from meridian.model import transformers
import tensorflow as tf


__all__ = [
    "MediaTensors",
    "RfTensors",
    "OrganicMediaTensors",
    "OrganicRfTensors",
    "build_media_tensors",
    "build_rf_tensors",
    "build_organic_media_tensors",
    "build_organic_rf_tensors",
]


@dataclasses.dataclass(frozen=True)
class MediaTensors:
  """Container for media tensors.

  Attributes:
    media: A tensor constructed from `InputData.media`.
    media_spend: A tensor constructed from `InputData.media_spend`.
    media_transformer: A `MediaTransformer` to scale media tensors using the
      model's media data.
    media_scaled: The media tensor normalized by population and by the median
      value.
    media_counterfactual: A tensor containing the media counterfactual values.
      If ROI priors are used, then the ROI of media channels is based on the
      difference in expected sales between the `media` tensor and this
      `media_counterfactual` tensor.
    media_counterfactual_scaled: A tensor containing the media counterfactual
      scaled values.
    media_spend_counterfactual: A tensor containing the media spend
      counterfactual values. If ROI priors are used, then the ROI of media
      channels is based on the spend difference between `media_spend` tensor and
      this `media_spend_counterfactual` tensor.
  """

  media: tf.Tensor | None = None
  media_spend: tf.Tensor | None = None
  media_transformer: transformers.MediaTransformer | None = None
  media_scaled: tf.Tensor | None = None
  media_counterfactual: tf.Tensor | None = None
  media_counterfactual_scaled: tf.Tensor | None = None
  media_spend_counterfactual: tf.Tensor | None = None


def build_media_tensors(
    input_data: data.InputData,
    model_spec: spec.ModelSpec,
) -> MediaTensors:
  """Derives a MediaTensors container from media values in given input data."""
  if input_data.media is None:
    return MediaTensors()

  # Derive and set media tensors from media values in the input data.
  media = tf.convert_to_tensor(input_data.media, dtype=tf.float32)
  media_spend = tf.convert_to_tensor(input_data.media_spend, dtype=tf.float32)
  media_transformer = transformers.MediaTransformer(
      media, tf.convert_to_tensor(input_data.population, dtype=tf.float32)
  )
  media_scaled = media_transformer.forward(media)

  # Derive counterfactual media tensors depending on whether mroi or roi priors
  # are used and whether roi_calibration_period is specified.
  if (
      model_spec.paid_media_prior_type
      == constants.PAID_MEDIA_PRIOR_TYPE_MROI
  ):
    factor = constants.MROI_FACTOR
  else:
    factor = 0

  if model_spec.roi_calibration_period is None:
    media_counterfactual = factor * media
    media_counterfactual_scaled = factor * media_scaled
    media_spend_counterfactual = factor * media_spend
  else:
    media_counterfactual = tf.where(
        model_spec.roi_calibration_period, factor * media, media
    )
    media_counterfactual_scaled = tf.where(
        model_spec.roi_calibration_period, factor * media_scaled, media_scaled
    )
    n_times = len(input_data.time)
    media_spend_counterfactual = tf.where(
        model_spec.roi_calibration_period[..., -n_times:, :],
        factor * media_spend,
        media_spend,
    )

  return MediaTensors(
      media=media,
      media_spend=media_spend,
      media_transformer=media_transformer,
      media_scaled=media_scaled,
      media_counterfactual=media_counterfactual,
      media_counterfactual_scaled=media_counterfactual_scaled,
      media_spend_counterfactual=media_spend_counterfactual,
  )


@dataclasses.dataclass(frozen=True)
class OrganicMediaTensors:
  """Container for organic media tensors.

  Attributes:
    organic_media: A tensor constructed from `InputData.organic_media`.
    organic_media_transformer: A `MediaTransformer` to scale media tensors using
      the model's organic media data.
    organic_media_scaled: The organic media tensor normalized by population and
      by the median value.
    organic_media_counterfactual: A tensor containing the organic media
      counterfactual values.
    organic_media_counterfactual_scaled: A tensor containing the organic media
      counterfactual scaled values.
  """

  organic_media: tf.Tensor | None = None
  organic_media_transformer: transformers.MediaTransformer | None = None
  organic_media_scaled: tf.Tensor | None = None
  organic_media_counterfactual: tf.Tensor | None = None
  organic_media_counterfactual_scaled: tf.Tensor | None = None


def build_organic_media_tensors(
    input_data: data.InputData,
) -> OrganicMediaTensors:
  """Derives a OrganicMediaTensors container from values in given input data."""
  if input_data.organic_media is None:
    return OrganicMediaTensors()

  # Derive and set media tensors from media values in the input data.
  organic_media = tf.convert_to_tensor(
      input_data.organic_media, dtype=tf.float32
  )
  organic_media_transformer = transformers.MediaTransformer(
      organic_media,
      tf.convert_to_tensor(input_data.population, dtype=tf.float32),
  )
  organic_media_scaled = organic_media_transformer.forward(organic_media)
  organic_media_counterfactual = tf.zeros_like(organic_media)
  organic_media_counterfactual_scaled = tf.zeros_like(organic_media_scaled)

  return OrganicMediaTensors(
      organic_media=organic_media,
      organic_media_transformer=organic_media_transformer,
      organic_media_scaled=organic_media_scaled,
      organic_media_counterfactual=organic_media_counterfactual,
      organic_media_counterfactual_scaled=organic_media_counterfactual_scaled,
  )


@dataclasses.dataclass(frozen=True)
class RfTensors:
  """Container for Reach and Frequency (RF) media tensors.

  Attributes:
    reach: A tensor constructed from `InputData.reach`.
    frequency: A tensor constructed from `InputData.frequency`.
    rf_spend: A tensor constructed from `InputData.rf_spend`.
    reach_transformer: A `MediaTransformer` to scale RF tensors using the
      model's RF data.
    reach_scaled: A reach tensor normalized by population and by the median
      value.
    reach_counterfactual: A reach tensor with media counterfactual values. If
      ROI priors are used, then the ROI of R&F channels is based on the
      difference in expected sales between the `reach` tensor and this
      `reach_counterfactual` tensor.
    reach_counterfactual_scaled: A reach tensor with media counterfactual scaled
      values.
    rf_spend_counterfactual: A reach tensor with media spend counterfactual
      values. If ROI priors are used, then the ROI of R&F channels is based on
      the spend difference between `rf_spend` tensor and this
      `rf_spend_counterfactual` tensor.
  """

  reach: tf.Tensor | None = None
  frequency: tf.Tensor | None = None
  rf_spend: tf.Tensor | None = None
  reach_transformer: transformers.MediaTransformer | None = None
  reach_scaled: tf.Tensor | None = None
  reach_counterfactual: tf.Tensor | None = None
  reach_counterfactual_scaled: tf.Tensor | None = None
  rf_spend_counterfactual: tf.Tensor | None = None


def build_rf_tensors(
    input_data: data.InputData,
    model_spec: spec.ModelSpec,
) -> RfTensors:
  """Derives an RfTensors container from RF media values in given input."""
  if input_data.reach is None:
    return RfTensors()

  reach = tf.convert_to_tensor(input_data.reach, dtype=tf.float32)
  frequency = tf.convert_to_tensor(input_data.frequency, dtype=tf.float32)
  rf_spend = tf.convert_to_tensor(input_data.rf_spend, dtype=tf.float32)
  reach_transformer = transformers.MediaTransformer(
      reach, tf.convert_to_tensor(input_data.population, dtype=tf.float32)
  )
  reach_scaled = reach_transformer.forward(reach)

  # marginal ROI by reach equals the ROI. The conversion between `beta_rf` and
  # `roi_rf` is the same, regardless of whether `roi_rf` represents ROI or
  # marginal ROI by reach.
  if model_spec.rf_roi_calibration_period is None:
    reach_counterfactual = tf.zeros_like(reach)
    reach_counterfactual_scaled = tf.zeros_like(reach_scaled)
    rf_spend_counterfactual = tf.zeros_like(rf_spend)
  else:
    reach_counterfactual = tf.where(
        model_spec.rf_roi_calibration_period, 0, reach
    )
    reach_counterfactual_scaled = tf.where(
        model_spec.rf_roi_calibration_period, 0, reach_scaled
    )
    n_times = len(input_data.time)
    rf_spend_counterfactual = tf.where(
        model_spec.rf_roi_calibration_period[..., -n_times:, :],
        0,
        rf_spend,
    )

  return RfTensors(
      reach=reach,
      frequency=frequency,
      rf_spend=rf_spend,
      reach_transformer=reach_transformer,
      reach_scaled=reach_scaled,
      reach_counterfactual=reach_counterfactual,
      reach_counterfactual_scaled=reach_counterfactual_scaled,
      rf_spend_counterfactual=rf_spend_counterfactual,
  )


@dataclasses.dataclass(frozen=True)
class OrganicRfTensors:
  """Container for Reach and Frequency (RF) organic media tensors.

  Attributes:
    organic_reach: A tensor constructed from `InputData.organic_reach`.
    organic_frequency: A tensor constructed from `InputData.organic_frequency`.
    organic_reach_transformer: A `MediaTransformer` to scale organic RF tensors
      using the model's organic RF data.
    organic_reach_scaled: An organic reach tensor normalized by population and
      by the median value.
    organic_reach_counterfactual: An organic reach tensor with media
      counterfactual values.
    organic_reach_counterfactual_scaled: An organic reach tensor with media
      counterfactual scaled values.
  """

  organic_reach: tf.Tensor | None = None
  organic_frequency: tf.Tensor | None = None
  organic_reach_transformer: transformers.MediaTransformer | None = None
  organic_reach_scaled: tf.Tensor | None = None
  organic_reach_counterfactual: tf.Tensor | None = None
  organic_reach_counterfactual_scaled: tf.Tensor | None = None


def build_organic_rf_tensors(
    input_data: data.InputData,
) -> OrganicRfTensors:
  """Derives an OrganicRfTensors container from values in given input."""
  if input_data.organic_reach is None:
    return OrganicRfTensors()

  organic_reach = tf.convert_to_tensor(
      input_data.organic_reach, dtype=tf.float32
  )
  organic_frequency = tf.convert_to_tensor(
      input_data.organic_frequency, dtype=tf.float32
  )
  organic_reach_transformer = transformers.MediaTransformer(
      organic_reach,
      tf.convert_to_tensor(input_data.population, dtype=tf.float32),
  )
  organic_reach_scaled = organic_reach_transformer.forward(organic_reach)
  organic_reach_counterfactual = tf.zeros_like(organic_reach)
  organic_reach_counterfactual_scaled = tf.zeros_like(organic_reach_scaled)

  return OrganicRfTensors(
      organic_reach=organic_reach,
      organic_frequency=organic_frequency,
      organic_reach_transformer=organic_reach_transformer,
      organic_reach_scaled=organic_reach_scaled,
      organic_reach_counterfactual=organic_reach_counterfactual,
      organic_reach_counterfactual_scaled=organic_reach_counterfactual_scaled,
  )



# Copyright 2024 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for sampling prior distributions in a Meridian model."""

from collections.abc import Mapping
from typing import TYPE_CHECKING

import arviz as az
from meridian import constants
import tensorflow as tf
import tensorflow_probability as tfp

if TYPE_CHECKING:
  from meridian.model import model  # pylint: disable=g-bad-import-order,g-import-not-at-top


__all__ = [
    "PriorDistributionSampler",
]


def _get_tau_g(
    tau_g_excl_baseline: tf.Tensor, baseline_geo_idx: int
) -> tfp.distributions.Distribution:
  """Computes `tau_g` from `tau_g_excl_baseline`.

  This function computes `tau_g` by inserting a column of zeros at the
  `baseline_geo` position in `tau_g_excl_baseline`.

  Args:
    tau_g_excl_baseline: A tensor of shape `[..., n_geos - 1]` for the
      user-defined dimensions of the `tau_g` parameter distribution.
    baseline_geo_idx: The index of the baseline geo to be set to zero.

  Returns:
    A tensor of shape `[..., n_geos]` with the final distribution of the `tau_g`
    parameter with zero at position `baseline_geo_idx` and matching
    `tau_g_excl_baseline` elsewhere.
  """
  rank = len(tau_g_excl_baseline.shape)
  shape = tau_g_excl_baseline.shape[:-1] + [1] if rank != 1 else 1
  tau_g = tf.concat(
      [
          tau_g_excl_baseline[..., :baseline_geo_idx],
          tf.zeros(shape, dtype=tau_g_excl_baseline.dtype),
          tau_g_excl_baseline[..., baseline_geo_idx:],
      ],
      axis=rank - 1,
  )
  return tfp.distributions.Deterministic(tau_g, name="tau_g")


class PriorDistributionSampler:
  """A callable that samples from a model spec's prior distributions."""

  def __init__(self, meridian: "model.Meridian"):
    self._meridian = meridian

  def get_roi_prior_beta_m_value(
      self,
      alpha_m: tf.Tensor,
      beta_gm_dev: tf.Tensor,
      ec_m: tf.Tensor,
      eta_m: tf.Tensor,
      roi_or_mroi_m: tf.Tensor,
      slope_m: tf.Tensor,
      media_transformed: tf.Tensor,
  ) -> tf.Tensor:
    """Returns a tensor to be used in `beta_m`."""
    mmm = self._meridian

    # The `roi_or_mroi_m` parameter represents either ROI or mROI. For reach &
    # frequency channels, marginal ROI priors are defined as "mROI by reach",
    # which is equivalent to ROI.
    media_spend = mmm.media_tensors.media_spend
    media_spend_counterfactual = mmm.media_tensors.media_spend_counterfactual
    media_counterfactual_scaled = mmm.media_tensors.media_counterfactual_scaled
    # If we got here, then we should already have media tensors derived from
    # non-None InputData.media data.
    assert media_spend is not None
    assert media_spend_counterfactual is not None
    assert media_counterfactual_scaled is not None

    # Use absolute value here because this difference will be negative for
    # marginal ROI priors.
    inc_revenue_m = roi_or_mroi_m * tf.reduce_sum(
        tf.abs(media_spend - media_spend_counterfactual),
        range(media_spend.ndim - 1),
    )

    if (
        mmm.model_spec.roi_calibration_period is None
        and mmm.model_spec.paid_media_prior_type
        == constants.PAID_MEDIA_PRIOR_TYPE_ROI
    ):
      # We can skip the adstock/hill computation step in this case.
      media_counterfactual_transformed = tf.zeros_like(media_transformed)
    else:
      media_counterfactual_transformed = mmm.adstock_hill_media(
          media=media_counterfactual_scaled,
          alpha=alpha_m,
          ec=ec_m,
          slope=slope_m,
      )

    revenue_per_kpi = mmm.revenue_per_kpi
    if mmm.input_data.revenue_per_kpi is None:
      revenue_per_kpi = tf.ones([mmm.n_geos, mmm.n_times], dtype=tf.float32)
    # Note: use absolute value here because this difference will be negative for
    # marginal ROI priors.
    media_contrib_gm = tf.einsum(
        "...gtm,g,,gt->...gm",
        tf.abs(media_transformed - media_counterfactual_transformed),
        mmm.population,
        mmm.kpi_transformer.population_scaled_stdev,
        revenue_per_kpi,
    )

    if mmm.media_effects_dist == constants.MEDIA_EFFECTS_NORMAL:
      media_contrib_m = tf.einsum("...gm->...m", media_contrib_gm)
      random_effect_m = tf.einsum(
          "...m,...gm,...gm->...m", eta_m, beta_gm_dev, media_contrib_gm
      )
      return (inc_revenue_m - random_effect_m) / media_contrib_m
    else:
      # For log_normal, beta_m and eta_m are not mean & std.
      # The parameterization is beta_gm ~ exp(beta_m + eta_m * N(0, 1)).
      random_effect_m = tf.einsum(
          "...gm,...gm->...m",
          tf.math.exp(beta_gm_dev * eta_m[..., tf.newaxis, :]),
          media_contrib_gm,
      )
    return tf.math.log(inc_revenue_m) - tf.math.log(random_effect_m)

  def get_roi_prior_beta_rf_value(
      self,
      alpha_rf: tf.Tensor,
      beta_grf_dev: tf.Tensor,
      ec_rf: tf.Tensor,
      eta_rf: tf.Tensor,
      roi_or_mroi_rf: tf.Tensor,
      slope_rf: tf.Tensor,
      rf_transformed: tf.Tensor,
  ) -> tf.Tensor:
    """Returns a tensor to be used in `beta_rf`."""
    mmm = self._meridian

    rf_spend = mmm.rf_tensors.rf_spend
    rf_spend_counterfactual = mmm.rf_tensors.rf_spend_counterfactual
    reach_counterfactual_scaled = mmm.rf_tensors.reach_counterfactual_scaled
    frequency = mmm.rf_tensors.frequency
    # If we got here, then we should already have RF media tensors derived from
    # non-None InputData.reach data.
    assert rf_spend is not None
    assert rf_spend_counterfactual is not None
    assert reach_counterfactual_scaled is not None
    assert frequency is not None

    inc_revenue_rf = roi_or_mroi_rf * tf.reduce_sum(
        rf_spend - rf_spend_counterfactual,
        range(rf_spend.ndim - 1),
    )
    if mmm.model_spec.rf_roi_calibration_period is not None:
      rf_counterfactual_transformed = mmm.adstock_hill_rf(
          reach=reach_counterfactual_scaled,
          frequency=frequency,
          alpha=alpha_rf,
          ec=ec_rf,
          slope=slope_rf,
      )
    else:
      rf_counterfactual_transformed = tf.zeros_like(rf_transformed)
    revenue_per_kpi = mmm.revenue_per_kpi
    if mmm.input_data.revenue_per_kpi is None:
      revenue_per_kpi = tf.ones([mmm.n_geos, mmm.n_times], dtype=tf.float32)

    media_contrib_grf = tf.einsum(
        "...gtm,g,,gt->...gm",
        rf_transformed - rf_counterfactual_transformed,
        mmm.population,
        mmm.kpi_transformer.population_scaled_stdev,
        revenue_per_kpi,
    )
    if mmm.media_effects_dist == constants.MEDIA_EFFECTS_NORMAL:
      media_contrib_rf = tf.einsum("...gm->...m", media_contrib_grf)
      random_effect_rf = tf.einsum(
          "...m,...gm,...gm->...m", eta_rf, beta_grf_dev, media_contrib_grf
      )
      return (inc_revenue_rf - random_effect_rf) / media_contrib_rf
    else:
      # For log_normal, beta_rf and eta_rf are not mean & std.
      # The parameterization is beta_grf ~ exp(beta_rf + eta_rf * N(0, 1)).
      random_effect_rf = tf.einsum(
          "...gm,...gm->...m",
          tf.math.exp(beta_grf_dev * eta_rf[..., tf.newaxis, :]),
          media_contrib_grf,
      )
      return tf.math.log(inc_revenue_rf) - tf.math.log(random_effect_rf)

  def _sample_media_priors(
      self,
      n_draws: int,
      seed: int | None = None,
  ) -> Mapping[str, tf.Tensor]:
    """Draws samples from the prior distributions of the media variables.

    Args:
      n_draws: Number of samples drawn from the prior distribution.
      seed: Used to set the seed for reproducible results. For more information,
        see [PRNGS and seeds]
        (https://github.com/tensorflow/probability/blob/main/PRNGS.md).

    Returns:
      A mapping of media parameter names to a tensor of shape `[n_draws, n_geos,
      n_media_channels]` or `[n_draws, n_media_channels]` containing the
      samples.
    """
    mmm = self._meridian

    prior = mmm.prior_broadcast
    sample_shape = [1, n_draws]
    sample_kwargs = {constants.SAMPLE_SHAPE: sample_shape, constants.SEED: seed}
    media_vars = {
        constants.ALPHA_M: prior.alpha_m.sample(**sample_kwargs),
        constants.EC_M: prior.ec_m.sample(**sample_kwargs),
        constants.ETA_M: prior.eta_m.sample(**sample_kwargs),
        constants.SLOPE_M: prior.slope_m.sample(**sample_kwargs),
    }
    beta_gm_dev = tfp.distributions.Sample(
        tfp.distributions.Normal(0, 1),
        [mmm.n_geos, mmm.n_media_channels],
        name=constants.BETA_GM_DEV,
    ).sample(**sample_kwargs)
    media_transformed = mmm.adstock_hill_media(
        media=mmm.media_tensors.media_scaled,
        alpha=media_vars[constants.ALPHA_M],
        ec=media_vars[constants.EC_M],
        slope=media_vars[constants.SLOPE_M],
    )

    prior_type = mmm.model_spec.paid_media_prior_type
    if prior_type == constants.PAID_MEDIA_PRIOR_TYPE_ROI:
      roi_m = prior.roi_m.sample(**sample_kwargs)
      beta_m_value = self.get_roi_prior_beta_m_value(
          beta_gm_dev=beta_gm_dev,
          media_transformed=media_transformed,
          roi_or_mroi_m=roi_m,
          **media_vars,
      )
      media_vars[constants.ROI_M] = roi_m
      media_vars[constants.BETA_M] = tfp.distributions.Deterministic(
          beta_m_value, name=constants.BETA_M
      ).sample()
    elif prior_type == constants.PAID_MEDIA_PRIOR_TYPE_MROI:
      mroi_m = prior.mroi_m.sample(**sample_kwargs)
      beta_m_value = self.get_roi_prior_beta_m_value(
          beta_gm_dev=beta_gm_dev,
          media_transformed=media_transformed,
          roi_or_mroi_m=mroi_m,
          **media_vars,
      )
      media_vars[constants.MROI_M] = mroi_m
      media_vars[constants.BETA_M] = tfp.distributions.Deterministic(
          beta_m_value, name=constants.BETA_M
      ).sample()
    else:
      media_vars[constants.BETA_M] = prior.beta_m.sample(**sample_kwargs)

    beta_eta_combined = (
        media_vars[constants.BETA_M][..., tf.newaxis, :]
        + media_vars[constants.ETA_M][..., tf.newaxis, :] * beta_gm_dev
    )
    beta_gm_value = (
        beta_eta_combined
        if mmm.media_effects_dist == constants.MEDIA_EFFECTS_NORMAL
        else tf.math.exp(beta_eta_combined)
    )
    media_vars[constants.BETA_GM] = tfp.distributions.Deterministic(
        beta_gm_value, name=constants.BETA_GM
    ).sample()

    return media_vars

  def _sample_rf_priors(
      self,
      n_draws: int,
      seed: int | None = None,
  ) -> Mapping[str, tf.Tensor]:
    """Draws samples from the prior distributions of the RF variables.

    Args:
      n_draws: Number of samples drawn from the prior distribution.
      seed: Used to set the seed for reproducible results. For more information,
        see [PRNGS and seeds]
        (https://github.com/tensorflow/probability/blob/main/PRNGS.md).

    Returns:
      A mapping of RF parameter names to a tensor of shape `[n_draws, n_geos,
      n_rf_channels]` or `[n_draws, n_rf_channels]` containing the samples.
    """
    mmm = self._meridian

    prior = mmm.prior_broadcast
    sample_shape = [1, n_draws]
    sample_kwargs = {constants.SAMPLE_SHAPE: sample_shape, constants.SEED: seed}
    rf_vars = {
        constants.ALPHA_RF: prior.alpha_rf.sample(**sample_kwargs),
        constants.EC_RF: prior.ec_rf.sample(**sample_kwargs),
        constants.ETA_RF: prior.eta_rf.sample(**sample_kwargs),
        constants.SLOPE_RF: prior.slope_rf.sample(**sample_kwargs),
    }
    beta_grf_dev = tfp.distributions.Sample(
        tfp.distributions.Normal(0, 1),
        [mmm.n_geos, mmm.n_rf_channels],
        name=constants.BETA_GRF_DEV,
    ).sample(**sample_kwargs)
    rf_transformed = mmm.adstock_hill_rf(
        reach=mmm.rf_tensors.reach_scaled,
        frequency=mmm.rf_tensors.frequency,
        alpha=rf_vars[constants.ALPHA_RF],
        ec=rf_vars[constants.EC_RF],
        slope=rf_vars[constants.SLOPE_RF],
    )

    prior_type = mmm.model_spec.paid_media_prior_type
    if prior_type == constants.PAID_MEDIA_PRIOR_TYPE_ROI:
      roi_rf = prior.roi_rf.sample(**sample_kwargs)
      beta_rf_value = self.get_roi_prior_beta_rf_value(
          beta_grf_dev=beta_grf_dev,
          rf_transformed=rf_transformed,
          roi_or_mroi_rf=roi_rf,
          **rf_vars,
      )
      rf_vars[constants.ROI_RF] = roi_rf
      rf_vars[constants.BETA_RF] = tfp.distributions.Deterministic(
          beta_rf_value,
          name=constants.BETA_RF,
      ).sample()
    elif prior_type == constants.PAID_MEDIA_PRIOR_TYPE_MROI:
      mroi_rf = prior.mroi_rf.sample(**sample_kwargs)
      beta_rf_value = self.get_roi_prior_beta_rf_value(
          beta_grf_dev=beta_grf_dev,
          rf_transformed=rf_transformed,
          roi_or_mroi_rf=mroi_rf,
          **rf_vars,
      )
      rf_vars[constants.MROI_RF] = mroi_rf
      rf_vars[constants.BETA_RF] = tfp.distributions.Deterministic(
          beta_rf_value,
          name=constants.BETA_RF,
      ).sample()
    else:
      rf_vars[constants.BETA_RF] = prior.beta_rf.sample(**sample_kwargs)

    beta_eta_combined = (
        rf_vars[constants.BETA_RF][..., tf.newaxis, :]
        + rf_vars[constants.ETA_RF][..., tf.newaxis, :] * beta_grf_dev
    )
    beta_grf_value = (
        beta_eta_combined
        if mmm.media_effects_dist == constants.MEDIA_EFFECTS_NORMAL
        else tf.math.exp(beta_eta_combined)
    )
    rf_vars[constants.BETA_GRF] = tfp.distributions.Deterministic(
        beta_grf_value, name=constants.BETA_GRF
    ).sample()

    return rf_vars

  def _sample_organic_media_priors(
      self,
      n_draws: int,
      seed: int | None = None,
  ) -> Mapping[str, tf.Tensor]:
    """Draws samples from the prior distributions of organic media variables.

    Args:
      n_draws: Number of samples drawn from the prior distribution.
      seed: Used to set the seed for reproducible results. For more information,
        see [PRNGS and seeds]
        (https://github.com/tensorflow/probability/blob/main/PRNGS.md).

    Returns:
      A mapping of organic media parameter names to a tensor of shape [n_draws,
      n_geos, n_organic_media_channels] or [n_draws, n_organic_media_channels]
      containing the samples.
    """
    mmm = self._meridian

    prior = mmm.prior_broadcast
    sample_shape = [1, n_draws]
    sample_kwargs = {constants.SAMPLE_SHAPE: sample_shape, constants.SEED: seed}
    organic_media_vars = {
        constants.ALPHA_OM: prior.alpha_om.sample(**sample_kwargs),
        constants.EC_OM: prior.ec_om.sample(**sample_kwargs),
        constants.ETA_OM: prior.eta_om.sample(**sample_kwargs),
        constants.SLOPE_OM: prior.slope_om.sample(**sample_kwargs),
    }
    beta_gom_dev = tfp.distributions.Sample(
        tfp.distributions.Normal(0, 1),
        [mmm.n_geos, mmm.n_organic_media_channels],
        name=constants.BETA_GOM_DEV,
    ).sample(**sample_kwargs)

    organic_media_vars[constants.BETA_OM] = prior.beta_om.sample(
        **sample_kwargs
    )

    beta_eta_combined = (
        organic_media_vars[constants.BETA_OM][..., tf.newaxis, :]
        + organic_media_vars[constants.ETA_OM][..., tf.newaxis, :]
        * beta_gom_dev
    )
    beta_gom_value = (
        beta_eta_combined
        if mmm.media_effects_dist == constants.MEDIA_EFFECTS_NORMAL
        else tf.math.exp(beta_eta_combined)
    )
    organic_media_vars[constants.BETA_GOM] = tfp.distributions.Deterministic(
        beta_gom_value, name=constants.BETA_GOM
    ).sample()

    return organic_media_vars

  def _sample_organic_rf_priors(
      self,
      n_draws: int,
      seed: int | None = None,
  ) -> Mapping[str, tf.Tensor]:
    """Draws samples from the prior distributions of the organic RF variables.

    Args:
      n_draws: Number of samples drawn from the prior distribution.
      seed: Used to set the seed for reproducible results. For more information,
        see [PRNGS and seeds]
        (https://github.com/tensorflow/probability/blob/main/PRNGS.md).

    Returns:
      A mapping of organic RF parameter names to a tensor of shape [n_draws,
      n_geos, n_organic_rf_channels] or [n_draws, n_organic_rf_channels]
      containing the samples.
    """
    mmm = self._meridian

    prior = mmm.prior_broadcast
    sample_shape = [1, n_draws]
    sample_kwargs = {constants.SAMPLE_SHAPE: sample_shape, constants.SEED: seed}
    organic_rf_vars = {
        constants.ALPHA_ORF: prior.alpha_orf.sample(**sample_kwargs),
        constants.EC_ORF: prior.ec_orf.sample(**sample_kwargs),
        constants.ETA_ORF: prior.eta_orf.sample(**sample_kwargs),
        constants.SLOPE_ORF: prior.slope_orf.sample(**sample_kwargs),
    }
    beta_gorf_dev = tfp.distributions.Sample(
        tfp.distributions.Normal(0, 1),
        [mmm.n_geos, mmm.n_organic_rf_channels],
        name=constants.BETA_GORF_DEV,
    ).sample(**sample_kwargs)

    organic_rf_vars[constants.BETA_ORF] = prior.beta_orf.sample(**sample_kwargs)

    beta_eta_combined = (
        organic_rf_vars[constants.BETA_ORF][..., tf.newaxis, :]
        + organic_rf_vars[constants.ETA_ORF][..., tf.newaxis, :] * beta_gorf_dev
    )
    beta_gorf_value = (
        beta_eta_combined
        if mmm.media_effects_dist == constants.MEDIA_EFFECTS_NORMAL
        else tf.math.exp(beta_eta_combined)
    )
    organic_rf_vars[constants.BETA_GORF] = tfp.distributions.Deterministic(
        beta_gorf_value, name=constants.BETA_GORF
    ).sample()

    return organic_rf_vars

  def _sample_non_media_treatments_priors(
      self,
      n_draws: int,
      seed: int | None = None,
  ) -> Mapping[str, tf.Tensor]:
    """Draws from the prior distributions of the non-media treatment variables.

    Args:
      n_draws: Number of samples drawn from the prior distribution.
      seed: Used to set the seed for reproducible results. For more information,
        see [PRNGS and seeds]
        (https://github.com/tensorflow/probability/blob/main/PRNGS.md).

    Returns:
      A mapping of non-media treatment parameter names to a tensor of shape
      [n_draws,
      n_geos, n_non_media_channels] or [n_draws, n_non_media_channels]
      containing the samples.
    """
    mmm = self._meridian

    prior = mmm.prior_broadcast
    sample_shape = [1, n_draws]
    sample_kwargs = {constants.SAMPLE_SHAPE: sample_shape, constants.SEED: seed}
    non_media_treatments_vars = {
        constants.GAMMA_N: prior.gamma_n.sample(**sample_kwargs),
        constants.XI_N: prior.xi_n.sample(**sample_kwargs),
    }
    gamma_gn_dev = tfp.distributions.Sample(
        tfp.distributions.Normal(0, 1),
        [mmm.n_geos, mmm.n_non_media_channels],
        name=constants.GAMMA_GN_DEV,
    ).sample(**sample_kwargs)
    non_media_treatments_vars[constants.GAMMA_GN] = (
        tfp.distributions.Deterministic(
            non_media_treatments_vars[constants.GAMMA_N][..., tf.newaxis, :]
            + non_media_treatments_vars[constants.XI_N][..., tf.newaxis, :]
            * gamma_gn_dev,
            name=constants.GAMMA_GN,
        ).sample()
    )
    return non_media_treatments_vars

  def _sample_prior(
      self,
      n_draws: int,
      seed: int | None = None,
  ) -> Mapping[str, tf.Tensor]:
    """Returns a mapping of prior parameters to tensors of the samples."""
    mmm = self._meridian

    # For stateful sampling, the random seed must be set to ensure that any
    # random numbers that are generated are deterministic.
    if seed is not None:
      tf.keras.utils.set_random_seed(1)

    prior = mmm.prior_broadcast
    sample_shape = [1, n_draws]
    sample_kwargs = {constants.SAMPLE_SHAPE: sample_shape, constants.SEED: seed}

    tau_g_excl_baseline = prior.tau_g_excl_baseline.sample(**sample_kwargs)
    base_vars = {
        constants.KNOT_VALUES: prior.knot_values.sample(**sample_kwargs),
        constants.GAMMA_C: prior.gamma_c.sample(**sample_kwargs),
        constants.XI_C: prior.xi_c.sample(**sample_kwargs),
        constants.SIGMA: prior.sigma.sample(**sample_kwargs),
        constants.TAU_G: _get_tau_g(
            tau_g_excl_baseline=tau_g_excl_baseline,
            baseline_geo_idx=mmm.baseline_geo_idx,
        ).sample(),
    }
    base_vars[constants.MU_T] = tfp.distributions.Deterministic(
        tf.einsum(
            "...k,kt->...t",
            base_vars[constants.KNOT_VALUES],
            tf.convert_to_tensor(mmm.knot_info.weights),
        ),
        name=constants.MU_T,
    ).sample()

    gamma_gc_dev = tfp.distributions.Sample(
        tfp.distributions.Normal(0, 1),
        [mmm.n_geos, mmm.n_controls],
        name=constants.GAMMA_GC_DEV,
    ).sample(**sample_kwargs)
    base_vars[constants.GAMMA_GC] = tfp.distributions.Deterministic(
        base_vars[constants.GAMMA_C][..., tf.newaxis, :]
        + base_vars[constants.XI_C][..., tf.newaxis, :] * gamma_gc_dev,
        name=constants.GAMMA_GC,
    ).sample()

    media_vars = (
        self._sample_media_priors(n_draws, seed)
        if mmm.media_tensors.media is not None
        else {}
    )
    rf_vars = (
        self._sample_rf_priors(n_draws, seed)
        if mmm.rf_tensors.reach is not None
        else {}
    )
    organic_media_vars = (
        self._sample_organic_media_priors(n_draws, seed)
        if mmm.organic_media_tensors.organic_media is not None
        else {}
    )
    organic_rf_vars = (
        self._sample_organic_rf_priors(n_draws, seed)
        if mmm.organic_rf_tensors.organic_reach is not None
        else {}
    )
    non_media_treatments_vars = (
        self._sample_non_media_treatments_priors(n_draws, seed)
        if mmm.non_media_treatments_scaled is not None
        else {}
    )

    return (
        base_vars
        | media_vars
        | rf_vars
        | organic_media_vars
        | organic_rf_vars
        | non_media_treatments_vars
    )

  def __call__(self, n_draws: int, seed: int | None = None) -> az.InferenceData:
    """Draws samples from prior distributions.

    Returns:
      An Arviz `InferenceData` object containing prior samples only.

    Args:
      n_draws: Number of samples drawn from the prior distribution.
      seed: Used to set the seed for reproducible results. For more information,
        see [PRNGS and seeds]
        (https://github.com/tensorflow/probability/blob/main/PRNGS.md).
    """
    prior_draws = self._sample_prior(n_draws, seed=seed)
    # Create Arviz InferenceData for prior draws.
    prior_coords = self._meridian.create_inference_data_coords(1, n_draws)
    prior_dims = self._meridian.create_inference_data_dims()
    return az.convert_to_inference_data(
        prior_draws, coords=prior_coords, dims=prior_dims, group=constants.PRIOR
    )



# Copyright 2024 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Constants shared across the Meridian library."""

import immutabledict

# HEX color map.
BLACK_100 = '#000000'

BLUE_700 = '#1967D2'
BLUE_600 = '#1A73E8'
BLUE_400 = '#669DF6'
BLUE_500 = '#4285F4'
BLUE_300 = '#8AB4F8'
BLUE_200 = '#AECBFA'

YELLOW_600 = '#F9AB00'
YELLOW_500 = '#FBBC04'

GREEN_700 = '#188038'
GREEN_500 = '#34A853'
GREEN_300 = '#81C995'

PURPLE_500 = '#AF5CF7'

RED_600 = '#D93025'
RED_300 = '#F28B82'
RED_200 = '#FC645B'

CYAN_500 = '#24C1E0'
CYAN_400 = '#4ECDE6'

GREY_900 = '#202124'
GREY_800 = '#3C4043'
GREY_700 = '#5F6368'
GREY_600 = '#C3C7C9'
GREY_300 = '#DADCE0'


# Example: "2024-01-09"
DATE_FORMAT = '%Y-%m-%d'

# Input data variables.
KPI = 'kpi'
REVENUE_PER_KPI = 'revenue_per_kpi'
MEDIA = 'media'
MEDIA_SPEND = 'media_spend'
CONTROLS = 'controls'
POPULATION = 'population'
REACH = 'reach'
FREQUENCY = 'frequency'
RF_SPEND = 'rf_spend'
ORGANIC_MEDIA = 'organic_media'
ORGANIC_REACH = 'organic_reach'
ORGANIC_FREQUENCY = 'organic_frequency'
NON_MEDIA_TREATMENTS = 'non_media_treatments'
REVENUE = 'revenue'
NON_REVENUE = 'non_revenue'
REQUIRED_INPUT_DATA_ARRAY_NAMES = (
    KPI,
    CONTROLS,
    POPULATION,
)
OPTIONAL_INPUT_DATA_ARRAY_NAMES = (
    REVENUE_PER_KPI,
    ORGANIC_MEDIA,
    ORGANIC_REACH,
    ORGANIC_FREQUENCY,
    NON_MEDIA_TREATMENTS,
)
MEDIA_INPUT_DATA_ARRAY_NAMES = (
    MEDIA,
    MEDIA_SPEND,
)
RF_INPUT_DATA_ARRAY_NAMES = (
    REACH,
    FREQUENCY,
    RF_SPEND,
)
POSSIBLE_INPUT_DATA_ARRAY_NAMES = (
    REQUIRED_INPUT_DATA_ARRAY_NAMES
    + OPTIONAL_INPUT_DATA_ARRAY_NAMES
    + MEDIA_INPUT_DATA_ARRAY_NAMES
    + RF_INPUT_DATA_ARRAY_NAMES
)

# Scaled input data variables.
MEDIA_SCALED = 'media_scaled'
REACH_SCALED = 'reach_scaled'
ORGANIC_MEDIA_SCALED = 'organic_media_scaled'
ORGANIC_REACH_SCALED = 'organic_reach_scaled'
NON_MEDIA_TREATMENTS_SCALED = 'non_media_treatments_scaled'
CONTROLS_SCALED = 'controls_scaled'

# Non-media treatments baseline value constants.
NON_MEDIA_BASELINE_MIN = 'min'
NON_MEDIA_BASELINE_MAX = 'max'
NON_MEDIA_BASELINE_VALUES = 'non_media_baseline_values'

# Input data coordinates.
GEO = 'geo'
TIME = 'time'
MEDIA_TIME = 'media_time'
MEDIA_CHANNEL = 'media_channel'
RF_CHANNEL = 'rf_channel'
CHANNEL = 'channel'
RF = 'rf'
ORGANIC_MEDIA_CHANNEL = 'organic_media_channel'
ORGANIC_RF_CHANNEL = 'organic_rf_channel'
NON_MEDIA_CHANNEL = 'non_media_channel'
CONTROL_VARIABLE = 'control_variable'
REQUIRED_INPUT_DATA_COORD_NAMES = (
    GEO,
    TIME,
    MEDIA_TIME,
    CONTROL_VARIABLE,
)
NON_PAID_MEDIA_INPUT_DATA_COORD_NAMES = (
    ORGANIC_MEDIA_CHANNEL,
    ORGANIC_RF_CHANNEL,
    NON_MEDIA_CHANNEL,
)
MEDIA_INPUT_DATA_COORD_NAMES = (MEDIA_CHANNEL,)
RF_INPUT_DATA_COORD_NAMES = (RF_CHANNEL,)
POSSIBLE_INPUT_DATA_COORD_NAMES = (
    REQUIRED_INPUT_DATA_COORD_NAMES
    + NON_PAID_MEDIA_INPUT_DATA_COORD_NAMES
    + MEDIA_INPUT_DATA_COORD_NAMES
    + RF_INPUT_DATA_COORD_NAMES
)
POSSIBLE_INPUT_DATA_COORDS_AND_ARRAYS_SET = frozenset(
    POSSIBLE_INPUT_DATA_COORD_NAMES + POSSIBLE_INPUT_DATA_ARRAY_NAMES
)


# National model constants.
NATIONAL = 'national'
NATIONAL_MODEL_DEFAULT_GEO_NAME = 'national_geo'
NATIONAL_MODEL_DEFAULT_POPULATION_VALUE = 1.0


# Media random effects distributions.
MEDIA_EFFECTS_NORMAL = 'normal'
MEDIA_EFFECTS_LOG_NORMAL = 'log_normal'
MEDIA_EFFECTS_DISTRIBUTIONS = frozenset(
    {MEDIA_EFFECTS_NORMAL, MEDIA_EFFECTS_LOG_NORMAL}
)


# Model spec variables.
PRIOR = 'prior'
MEDIA_EFFECTS_DIST = 'media_effects_dist'
HILL_BEFORE_ADSTOCK = 'hill_before_adstock'
ADSTOCK_MEMORY_OPTIMIZED = 'adstock_memory_optimized'
MAX_LAG = 'max_lag'
UNIQUE_SIGMA_FOR_EACH_GEO = 'unique_sigma_for_each_geo'
ROI_CALIBRATION_PERIOD = 'roi_calibration_period'
RF_ROI_CALIBRATION_PERIOD = 'rf_roi_calibration_period'
KNOTS = 'knots'
BASELINE_GEO = 'baseline_geo'

# Media prior types.
PAID_MEDIA_PRIOR_TYPE_ROI = 'roi'
PAID_MEDIA_PRIOR_TYPE_MROI = 'mroi'
PAID_MEDIA_PRIOR_TYPE_COEFFICIENT = 'coefficient'
PAID_MEDIA_PRIOR_TYPES = frozenset({
    PAID_MEDIA_PRIOR_TYPE_ROI,
    PAID_MEDIA_PRIOR_TYPE_MROI,
    PAID_MEDIA_PRIOR_TYPE_COEFFICIENT,
})
PAID_MEDIA_ROI_PRIOR_TYPES = frozenset(
    {PAID_MEDIA_PRIOR_TYPE_ROI, PAID_MEDIA_PRIOR_TYPE_MROI}
)
# Represents a 1% increase in spend.
MROI_FACTOR = 1.01

NATIONAL_MODEL_SPEC_ARGS = immutabledict.immutabledict(
    {MEDIA_EFFECTS_DIST: MEDIA_EFFECTS_NORMAL, UNIQUE_SIGMA_FOR_EACH_GEO: False}
)

NATIONAL_ANALYZER_PARAMETERS_DEFAULTS = immutabledict.immutabledict(
    {'aggregate_geos': True, 'geos_to_include': None}
)


# Inference data coordinates.
CHAIN = 'chain'
DRAW = 'draw'
KNOTS = 'knots'
SIGMA_DIM = 'sigma_dim'


# Model parameters.
PARAMETER = 'parameter'
KNOT_VALUES = 'knot_values'
MU_T = 'mu_t'
ROI_M = 'roi_m'
ROI_RF = 'roi_rf'
MROI_M = 'mroi_m'
MROI_RF = 'mroi_rf'
GAMMA_C = 'gamma_c'
GAMMA_N = 'gamma_n'
XI_C = 'xi_c'
XI_N = 'xi_n'
ALPHA_M = 'alpha_m'
ALPHA_RF = 'alpha_rf'
EC_M = 'ec_m'
EC_RF = 'ec_rf'
SLOPE_M = 'slope_m'
SLOPE_RF = 'slope_rf'
ETA_M = 'eta_m'
ETA_RF = 'eta_rf'
BETA_M = 'beta_m'
BETA_RF = 'beta_rf'
BETA_GM = 'beta_gm'
BETA_GRF = 'beta_grf'
BETA_OM = 'beta_om'
BETA_ORF = 'beta_orf'
ETA_OM = 'eta_om'
ETA_ORF = 'eta_orf'
ALPHA_OM = 'alpha_om'
ALPHA_ORF = 'alpha_orf'
EC_OM = 'ec_om'
EC_ORF = 'ec_orf'
SLOPE_OM = 'slope_om'
SLOPE_ORF = 'slope_orf'
BETA_GOM = 'beta_gom'
BETA_GORF = 'beta_gorf'
SIGMA = 'sigma'
TAU_G = 'tau_g'
TAU_G_EXCL_BASELINE = 'tau_g_excl_baseline'
GAMMA_GC = 'gamma_gc'
GAMMA_GN = 'gamma_gn'
BETA_GM_DEV = 'beta_gm_dev'
BETA_GRF_DEV = 'beta_grf_dev'
BETA_GOM_DEV = 'beta_gom_dev'
BETA_GORF_DEV = 'beta_gorf_dev'
GAMMA_GC_DEV = 'gamma_gc_dev'
GAMMA_GN_DEV = 'gamma_gn_dev'
COMMON_PARAMETER_NAMES = (
    KNOT_VALUES,
    MU_T,
    GAMMA_C,
    XI_C,
    SIGMA,
    TAU_G,
    GAMMA_GC,
)
# These constants are only used in unit tests for mocking default inference data
# which doesn't include MROI priors.
MEDIA_PARAMETER_NAMES = (
    ROI_M,
    ALPHA_M,
    EC_M,
    SLOPE_M,
    ETA_M,
    BETA_M,
    BETA_GM,
)
RF_PARAMETER_NAMES = (
    ROI_RF,
    ALPHA_RF,
    EC_RF,
    SLOPE_RF,
    ETA_RF,
    BETA_RF,
    BETA_GRF,
)
ORGANIC_MEDIA_PARAMETERS = (
    BETA_OM,
    ETA_OM,
    ALPHA_OM,
    EC_OM,
    SLOPE_OM,
)
ORGANIC_RF_PARAMETERS = (
    BETA_ORF,
    ETA_ORF,
    ALPHA_ORF,
    EC_ORF,
    SLOPE_ORF,
)
NON_MEDIA_PARAMETERS = (
    GAMMA_N,
    XI_N,
)

KNOTS_PARAMETERS = (KNOT_VALUES,)
MEDIA_PARAMETERS = (ETA_M, BETA_M, ALPHA_M, EC_M, SLOPE_M, ROI_M, MROI_M)
RF_PARAMETERS = (ETA_RF, BETA_RF, ALPHA_RF, EC_RF, SLOPE_RF, ROI_RF, MROI_RF)
CONTROL_PARAMETERS = (GAMMA_C, XI_C)
SIGMA_PARAMETERS = (SIGMA,)
GEO_PARAMETERS = (
    # TAU_G in InferenceData is derived from TAU_G_EXCL_BASELINE in the priors.
    TAU_G,
)
TIME_PARAMETERS = (
    # MU_T in InferenceData is derived from KNOT_VALUES in the priors.
    MU_T,
)
GEO_MEDIA_PARAMETERS = (BETA_GM,)
GEO_RF_PARAMETERS = (BETA_GRF,)
GEO_CONTROL_PARAMETERS = (GAMMA_GC,)
GEO_NON_MEDIA_PARAMETERS = (GAMMA_GN,)

ALL_PRIOR_DISTRIBUTION_PARAMETERS = (
    *KNOTS_PARAMETERS,
    *MEDIA_PARAMETERS,
    *RF_PARAMETERS,
    *ORGANIC_MEDIA_PARAMETERS,
    *ORGANIC_RF_PARAMETERS,
    *NON_MEDIA_PARAMETERS,
    *CONTROL_PARAMETERS,
    *SIGMA_PARAMETERS,
    TAU_G_EXCL_BASELINE,
    *TIME_PARAMETERS,
)

UNSAVED_PARAMETERS = (
    BETA_GM_DEV,
    BETA_GRF_DEV,
    BETA_GOM_DEV,
    BETA_GORF_DEV,
    GAMMA_GC_DEV,
    GAMMA_GN_DEV,
    TAU_G_EXCL_BASELINE,  # Used to derive TAU_G.
)
IGNORED_PRIORS = immutabledict.immutabledict({
    PAID_MEDIA_PRIOR_TYPE_ROI: (BETA_M, BETA_RF, MROI_M, MROI_RF),
    PAID_MEDIA_PRIOR_TYPE_MROI: (BETA_M, BETA_RF, ROI_M, ROI_RF),
    PAID_MEDIA_PRIOR_TYPE_COEFFICIENT: (ROI_M, ROI_RF, MROI_M, MROI_RF),
})

# Inference data dimensions.
INFERENCE_DIMS = immutabledict.immutabledict(
    {
        MU_T: (TIME,),
        KNOT_VALUES: (KNOTS,),
        TAU_G: (GEO,),
        BETA_GM: (GEO, MEDIA_CHANNEL),
        BETA_GRF: (GEO, RF_CHANNEL),
        BETA_GOM: (GEO, ORGANIC_MEDIA_CHANNEL),
        BETA_GORF: (GEO, ORGANIC_RF_CHANNEL),
        GAMMA_GC: (GEO, CONTROL_VARIABLE),
        GAMMA_GN: (GEO, NON_MEDIA_CHANNEL),
    }
    | {param: (CONTROL_VARIABLE,) for param in CONTROL_PARAMETERS}
    | {param: (NON_MEDIA_CHANNEL,) for param in NON_MEDIA_PARAMETERS}
    | {param: (MEDIA_CHANNEL,) for param in MEDIA_PARAMETERS}
    | {param: (RF_CHANNEL,) for param in RF_PARAMETERS}
    | {param: (ORGANIC_MEDIA_CHANNEL,) for param in ORGANIC_MEDIA_PARAMETERS}
    | {param: (ORGANIC_RF_CHANNEL,) for param in ORGANIC_RF_PARAMETERS}
)

IGNORED_TRACE_METRICS = ('variance_scaling',)

STEP_SIZE = 'step_size'
TUNE = 'tune'
TARGET_LOG_PROBABILITY_TF = 'target_log_prob'
TARGET_LOG_PROBABILITY_ARVIZ = 'lp'
DIVERGING = 'diverging'
ACCEPT_RATIO = 'accept_ratio'
N_STEPS = 'n_steps'
SAMPLE_SHAPE = 'sample_shape'
SEED = 'seed'

SAMPLE_STATS_METRICS = immutabledict.immutabledict({
    STEP_SIZE: STEP_SIZE,
    TARGET_LOG_PROBABILITY_TF: TARGET_LOG_PROBABILITY_ARVIZ,
    DIVERGING: DIVERGING,
    N_STEPS: N_STEPS,
})


# Adstock hill functions.
ADSTOCK_HILL_FUNCTIONS = frozenset({
    'adstock_memory_optimized',
    'adstock_speed_optimized',
    'hill',
})


# Distribution constants.
DISTRIBUTION = 'distribution'
DISTRIBUTION_TYPE = 'distribution_type'
PRIOR = 'prior'
POSTERIOR = 'posterior'
# Prior mean proportion of KPI incremental due to all media.
P_MEAN = 0.4
# Prior standard deviation proportion of KPI incremental to all media.
P_SD = 0.2


# Model metrics.
METRIC = 'metric'
MEAN = 'mean'
MEDIAN = 'median'
CI_LO = 'ci_lo'
CI_HI = 'ci_hi'
RHAT = 'rhat'
PCT = 'pct'
CONFIDENCE_LEVEL = 'confidence_level'


# Model fit types.
TYPE = 'type'
EXPECTED = 'expected'
ACTUAL = 'actual'
BASELINE = 'baseline'

# Model fit filtering.
GEO_GRANULARITY = 'geo_granularity'
EVALUATION_SET_VAR = 'evaluation_set'
VALUE = 'value'

# Predictive accuracy metrics.
R_SQUARED = 'R_Squared'
MAPE = 'MAPE'
WMAPE = 'wMAPE'

# Data splitting types.
TRAIN = 'Train'
TEST = 'Test'
ALL_DATA = 'All Data'

# Model fit sets.
EVALUATION_SET = (
    TRAIN,
    TEST,
    ALL_DATA,
)
GEO_GRANULARITY_SET = (GEO, NATIONAL)
METRICS_SET = (R_SQUARED, MAPE, WMAPE)
COLUMN_VAR_SET = (None, METRIC, GEO_GRANULARITY, EVALUATION_SET)

# Media effects metrics.
SPEND_MULTIPLIER = 'spend_multiplier'
TIME_UNITS = 'time_units'
MEDIA_UNITS = 'media_units'
HILL_SATURATION_LEVEL = 'hill_saturation_level'
OPTIMAL_FREQUENCY = 'optimal_frequency'
EXPECTED_ROI = 'expected_roi'
EFFECT = 'effect'

# Media effects helper values
MEDIA_UNITS_COUNT_HISTOGRAM = 'media_units_count_histogram'
SCALED_COUNT_HISTOGRAM = 'scaled_count_histogram'
START_INTERVAL_HISTOGRAM = 'start_interval_histogram'
END_INTERVAL_HISTOGRAM = 'end_interval_histogram'
COUNT_HISTOGRAM = 'count_histogram'
CHANNEL_TYPE = 'channel_type'
IS_INT_TIME_UNIT = 'is_int_time_unit'
OPTIMAL_FREQ = 'optimal_freq'
CURRENT_SPEND = 'current_spend'

# Media summary metrics.
SPEND = 'spend'
IMPRESSIONS = 'impressions'
ROI = 'roi'
OPTIMIZED_ROI = 'optimized_roi'
MROI = 'mroi'
OPTIMIZED_MROI_BY_REACH = 'optimized_mroi_by_reach'
OPTIMIZED_MROI_BY_FREQUENCY = 'optimized_mroi_by_frequency'
CPIK = 'cpik'
OPTIMIZED_CPIK = 'optimized_cpik'
ROI_SCALED = 'roi_scaled'
OUTCOME = 'outcome'
INCREMENTAL_OUTCOME = 'incremental_outcome'
BASELINE_OUTCOME = 'baseline_outcome'
OPTIMIZED_INCREMENTAL_OUTCOME = 'optimized_incremental_outcome'
EFFECTIVENESS = 'effectiveness'
OPTIMIZED_EFFECTIVENESS = 'optimized_effectiveness'
PCT_OF_IMPRESSIONS = 'pct_of_impressions'
PCT_OF_SPEND = 'pct_of_spend'
CPM = 'cpm'
PCT_OF_CONTRIBUTION = 'pct_of_contribution'
OPTIMIZED_PCT_OF_CONTRIBUTION = 'optimized_pct_of_contribution'
BUDGET = 'budget'
PROFIT = 'profit'
IS_REVENUE_KPI = 'is_revenue_kpi'
TOTAL_INCREMENTAL_OUTCOME = 'total_incremental_outcome'
TOTAL_ROI = 'total_roi'
TOTAL_CPIK = 'total_cpik'
USE_KPI = 'use_kpi'
USE_POSTERIOR = 'use_posterior'

# R-hat summary metrics.
PARAM = 'param'
N_PARAMS = 'n_params'
AVG_RHAT = 'avg_rhat'
MAX_RHAT = 'max_rhat'
PERCENT_BAD_RHAT = 'percent_bad_rhat'
ROW_IDX_BAD_RHAT = 'row_idx_bad_rhat'
COL_IDX_BAD_RHAT = 'col_idx_bad_rhat'


# Analyzer Parameters.
NEW_MEDIA = 'new_media'
NEW_MEDIA_SPEND = 'new_media_spend'
NEW_REACH = 'new_reach'
NEW_FREQUENCY = 'new_frequency'
NEW_RF_SPEND = 'new_rf_spend'
NEW_ORGANIC_MEDIA = 'new_organic_media'
NEW_ORGANIC_REACH = 'new_organic_reach'
NEW_ORGANIC_FREQUENCY = 'new_organic_frequency'
NEW_NON_MEDIA_TREATMENTS = 'new_non_media_treatments'
NEW_CONTROLS = 'new_controls'
NEW_DATA = 'new_data'


# Media types.
ALL_CHANNELS = 'All Channels'

# Optimization constants.
OPTIMIZED = 'optimized'
NON_OPTIMIZED = 'non_optimized'
SPEND_CONSTRAINT = 'spend_constraint'
SPEND_LEVEL = 'spend_level'
LOWER_BOUND = 'lower_bound'
UPPER_BOUND = 'upper_bound'
SPEND_GRID = 'spend_grid'
SPEND_STEP_SIZE = 'spend_step_size'
INCREMENTAL_OUTCOME_GRID = 'incremental_outcome_grid'
GRID_SPEND_INDEX = 'grid_spend_index'
USE_HISTORICAL_BUDGET = 'use_historical_budget'


# Optimization constraints.
FIXED_BUDGET = 'fixed_budget'
TARGET_ROI = 'target_roi'
TARGET_MROI = 'target_mroi'
SPEND_CONSTRAINT_DEFAULT_FIXED_BUDGET = 0.3
SPEND_CONSTRAINT_DEFAULT_FLEXIBLE_BUDGET = 1.0


# Plot constants.
BAR_SIZE = 42
PADDING_10 = 10
PADDING_20 = 20
AXIS_FONT_SIZE = 12
TEXT_FONT_SIZE = 14
TITLE_FONT_SIZE = 18
SCALED_PADDING = 0.2
CORNER_RADIUS = 2
STROKE_DASH = (4, 2)
POINT_SIZE = 80
INDEPENDENT = 'independent'
RESPONSE_CURVE_STEP_SIZE = 0.01


# Font names.
FONT_ROBOTO = 'Roboto'
FONT_GOOGLE_SANS_DISPLAY = 'Google Sans Display'

# Default confidence level for the analysis.
DEFAULT_CONFIDENCE_LEVEL = 0.9

# Default number of max draws per chain in Analyzer.expected_outcome()
DEFAULT_BATCH_SIZE = 100


# Optimization constants.
CHAINS_DIMENSION = 0
DRAWS_DIMENSION = 1
SPEND_CONSTRAINT_PADDING = 0.5


# Range constants for defining the step_size for the x-axis in select plots.
ADSTOCK_DECAY_DEFAULT_STEPS_PER_TIME_PERIOD = 5
ADSTOCK_DECAY_MAX_TOTAL_STEPS = 1000
HILL_NUM_STEPS = 500

# Summary template params.
START_DATE = 'start_date'
END_DATE = 'end_date'
CARD_INSIGHTS = 'insights'
CARD_CHARTS = 'charts'
CARD_STATS = 'stats'

# VegaLite common params.
VEGALITE_FACET_DEFAULT_WIDTH = 400
